import torch

from .base import Transform


class PadFront(Transform):
    def __init__(self, length):
        self.length = length

    def __call__(self, ts):
        return torch.cat([torch.zeros(self.length, ts.shape[1]), ts], dim=0)

    def __repr__(self):
        return self.__class__.__name__ + "(length={0})".format(self.length)


class PadBack(Transform):
    def __init__(self, length):
        self.length = length

    def __call__(self, ts):
        return torch.cat([ts, torch.zeros(self.length, ts.shape[1])], dim=0)

    def __repr__(self):
        return self.__class__.__name__ + "(length={0})".format(self.length)



import os
import shutil
import tempfile
import urllib.request
import zipfile
from pathlib import Path

from sktime.datasets._data_io import _list_available_datasets
from .typing import AnyPath


from pathlib import Path
import zipfile


def _list_available_datasets(data_dir: AnyPath) -> list[str]:
    datasets = []
    for name in os.listdir(data_dir):
        sub_dir = os.path.join(data_dir, name)
        if os.path.isdir(sub_dir):
            all_files = os.listdir(sub_dir)
            if name + "_TRAIN.ts" in all_files and name + "_TEST.ts" in all_files:
                datasets.append(name)
    return datasets


def download_uea_ucr(dataset_name: str, extract_path: str | Path | None = None) -> bool:
    """
    Downloads the UEA/UCR time series dataset from sktime to the specified path.

    Parameters:
    dataset_name (str): The name of the dataset to be downloaded.
    extract_path (str | Path | None, default=None) : The path to extract the dataset. If None, extract to `.cache/data` directory.

    Returns:
    bool: True if the dataset was already present, False otherwise.

    Raises:
    ValueError: If the dataset is not available on the given path or on 'https://timeseriesclassification.com/'.
    """
    # Download UCR/UEA archive from sktime
    if extract_path is not None:
        local_dirname = Path(extract_path)
    else:
        local_dirname = Path(".cache/data")

    full_path = local_dirname / dataset_name

    full_path.mkdir(parents=True, exist_ok=True)

    if dataset_name not in _list_available_datasets(local_dirname):
        # Dataset is not already present in the datasets directory provided.
        # If it is not there, download and install it.
        url = f"https://timeseriesclassification.com/Downloads/{dataset_name}.zip"
        try:
            download_and_unzip_dataset(url, full_path)
            # If zip contains folder name as root, move contents up one level
            if len(os.listdir(full_path)) == 1:
                subfolder = os.listdir(full_path)[0]
                subfolder_path = full_path / subfolder
                for file in os.listdir(subfolder_path):
                    shutil.move(subfolder_path / file, full_path)
                os.rmdir(subfolder_path)
        except zipfile.BadZipFile as e:
            raise ValueError(
                f"Invalid dataset name '{dataset_name}' is not available on extract path '{extract_path}'. "
                f"Nor is it available on 'https://timeseriesclassification.com/'."
            ) from e
        return False
    else:
        return True


def download_and_unzip_dataset(url: str, path: AnyPath) -> None:
    with tempfile.NamedTemporaryFile() as zipped_file:
        with urllib.request.urlopen(url) as response:
            shutil.copyfileobj(response, zipped_file)

        with zipfile.ZipFile(zipped_file) as zf:
            zf.extractall(path)



from pathlib import Path

import torch
from dataclasses import dataclass
from pathlib import Path


def swap_batch_seq_collate_fn(data):
    xs, ys = zip(*data)

    xs = torch.stack(xs).permute(1, 0, 2)

    return xs, torch.stack(ys)


@dataclass
class TSInfo:
    num_classes: int
    series_length: int
    equal_length: bool
    dimensions: int
    univariate: bool
    missing: bool


def parse_ts(path: Path):
    num_classes = 0
    series_length = -1
    equal_length = False
    dimensions = -1
    univariate = False
    missing = False

    with path.open("r") as f:
        for line in f:
            if "@data" in line:
                break

            split = line.strip().split(" ")
            if split[0] == "@classLabel":
                if split[1] == "true":
                    num_classes = len(split[2:])
            elif split[0] == "@seriesLength" and split[1].isdigit():
                series_length = int(split[1])
            elif split[0] == "@equalLength" and split[1] == "true":
                equal_length = True
            elif split[0] == "@missing" and split[1] == "true":
                missing = True
            elif split[0] == "@dimensions" and split[1].isdigit():
                dimensions = int(split[1])
            elif split[0] == "@univariate" and split[1] == "true":
                univariate = True
                dimensions = 1
    return TSInfo(
        num_classes, series_length, equal_length, dimensions, univariate, missing
    )



from os import PathLike
from typing import TypeAlias
from enum import Enum

AnyPath: TypeAlias = str | bytes | PathLike


class DatasetSplit(Enum):
    TRAIN = "train"
    VAL = "val"
    TEST = "test"



import math
from pathlib import Path
from typing import Literal
from lightning import LightningDataModule

from ..typing import DatasetSplit

from ..datasets import UCRUEADataset
from ..download import download_uea_ucr
from torch.utils.data import random_split, DataLoader

from ..utils import swap_batch_seq_collate_fn

from ..transforms import Transform


class UCRUEADataModule(LightningDataModule):
    """UCR/UEA dataset datamodule

    Args:
        name : str
            Name of the dataset.
        split_ratio : float | tuple[float, float], optional (default: (0.75, 0.15))
            Ratio of train, val, test split of the entire dataset.
            If a tuple of two floats is passed, we define the ratio of train and validation splits on the entire dataset i.e.
            TRAIN_SPLIT + TEST_SPLIT of the original UEA_UCR dataset. For instance,
            if the TRAIN_SPLIT has a length of 10 and TEST_SPLIT has a length of 2, and we choose (0.75, 0.15),
            this will result in 75% train, 15% validation, 10% test, and thus 9 train, 2 val, 1 test.
            If a float is passed, we define the ratio of train and validation splits on the TRAIN_SPLIT of the original UEA_UCR dataset, and the TEST_SPLIT will remain untouched.
            For instance, if the TRAIN_SPLIT has a length of 10 and TEST_SPLIT has a length of 2, and we choose 0.75,
            this will result in 75% train, 25% validation, 10% test, and thus 7 train, 3 val, 2 test.
        batch_size : int, optional (default: 32)
            The batch size to use for the dataloaders.
        transform : Transform or None, optional (default: None)
            The transform to apply to the dataset.

    Attributes:
        cache_dir : pathlib.Path
            Directory where the data will be downloaded and stored.
    """

    def __init__(
        self,
        name: str,
        split_ratio: tuple[float, float] | float = (0.75, 0.15),
        batch_size: int = 32,
        transform: Transform | None = None,
    ):
        super().__init__()

        self.batch_size = batch_size
        self.cache_dir = Path(".cache") / "data"
        self.split_ratio = split_ratio
        self.name = name
        self.transform = transform

    def prepare_data(self):
        download_uea_ucr(self.name, self.cache_dir)

    def label_from_index(self, index: int) -> str:
        assert self.__label_from_index is not None, "You need to call setup first"
        return self.__label_from_index(index)

    @property
    def num_classes(self) -> int:
        assert self.__num_classes is not None, "You need to call setup first"
        return self.__num_classes

    @property
    def dimensions(self) -> int:
        assert self.__dimensions is not None, "You need to call setup first"
        return self.__dimensions

    @property
    def series_length(self) -> int:
        assert self.__series_length is not None, "You need to call setup first"
        return self.__series_length

    @property
    def equal_length(self) -> bool:
        assert self.__equal_length is not None, "You need to call setup first"
        return self.__equal_length

    @property
    def univariate(self) -> bool:
        assert self.__univariate is not None, "You need to call setup first"
        return self.__univariate

    @property
    def missing(self) -> bool:
        assert self.__missing is not None, "You need to call setup first"
        return self.__missing

    def setup(self, stage=None):
        if isinstance(self.split_ratio, float):
            train_split_dataset = UCRUEADataset(
                self.name,
                self.cache_dir,
                split=DatasetSplit.TRAIN,
                transform=self.transform,
            )
            val_size = round(1 - self.split_ratio, 2)
            self.train_dataset, self.val_dataset = random_split(
                train_split_dataset, [self.split_ratio, val_size]
            )
            self.test_dataset = UCRUEADataset(
                self.name,
                self.cache_dir,
                split=DatasetSplit.TEST,
                transform=self.transform,
            )

            self.__label_from_index = train_split_dataset.label_from_index
            self.__num_classes = train_split_dataset.num_classes
            self.__dimensions = train_split_dataset.dimensions
            self.__series_length = train_split_dataset.series_length
            self.__equal_length = train_split_dataset.equal_length
            self.__univariate = train_split_dataset.univariate
            self.__missing = train_split_dataset.missing
        else:
            dataset = UCRUEADataset(
                ds_name=self.name, path=self.cache_dir, transform=self.transform
            )
            self.__label_from_index = dataset.label_from_index
            self.__num_classes = dataset.num_classes
            self.__dimensions = dataset.dimensions
            self.__series_length = dataset.series_length
            self.__equal_length = dataset.equal_length
            self.__univariate = dataset.univariate
            self.__missing = dataset.missing

            # split dataset
            test_size = round(1 - self.split_ratio[0] - self.split_ratio[1], 2)
            if math.isclose(test_size, 0.0):
                self.train_dataset, self.val_dataset = random_split(
                    dataset, [self.split_ratio[0], self.split_ratio[1]]
                )
            else:
                self.train_dataset, self.val_dataset, self.test_dataset = random_split(
                    dataset, [self.split_ratio[0], self.split_ratio[1], test_size]
                )

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            collate_fn=swap_batch_seq_collate_fn,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            collate_fn=swap_batch_seq_collate_fn,
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            collate_fn=swap_batch_seq_collate_fn,
        )



from pathlib import Path
import pandas as pd
from sktime.datasets._data_io import _load_provided_dataset
import numpy as np
import torch
from torch.utils.data import Dataset
from ..errors import MissingValueError
from ..typing import DatasetSplit
from ..transforms import Transform
from ..utils import parse_ts


class UCRUEADataset(Dataset):
    """UCR/UEA dataset.

    Args:
        ds_name (str): Name of the dataset.
        path (Path): Path to the dataset storage.
        split (DatasetSplit, optional): Dataset split to use. Defaults to None.
        transform (Transform, optional): Transform to apply to the data. Defaults to None.
        torchchronos_cache (bool, optional): Whether to cache the data in torchchronos format. Defaults to True.
            This allows for much faster loading of the data once it has been loaded once.
            It is recommended to leave this as True unless space is an issue.
            The file is not compressed to allow for faster loading.
    """

    def __init__(
        self,
        ds_name: str,
        path: Path,
        split: DatasetSplit | None = None,
        transform: Transform | None = None,
        torchchronos_cache: bool = True,
        raise_on_missing: bool = True,
    ) -> None:
        super().__init__()

        self.transform = transform
        match split:
            case None:
                split_val = None
            case DatasetSplit.TRAIN:
                split_val = "TRAIN"
            case DatasetSplit.VAL:
                raise ValueError("UCR/UEA datasets do not have a validation split")
            case DatasetSplit.TEST:
                split_val = "TEST"

        tc_cache_dir = path / ds_name
        tc_cache_path = tc_cache_dir / f"{split_val}.npz"
        if torchchronos_cache and tc_cache_path.exists():
            # memory-mapping is slightly faster
            data = np.load(tc_cache_path, mmap_mode="r")
            self.xs, self.ys = data["xs"], data["ys"]
        else:
            self.xs, self.ys = _load_provided_dataset(
                ds_name,
                split=split_val,
                return_type="numpy3d",
                local_module=path.parent,
                local_dirname=path.stem,
            )
            if torchchronos_cache:
                tc_cache_dir.mkdir(parents=True, exist_ok=True)
                # Using savez instead of savez_compressed because the latter is much slower
                np.savez(tc_cache_path, xs=self.xs, ys=self.ys)

        self.xs = torch.tensor(self.xs, dtype=torch.float32).transpose(1, 2)
        if self.transform is not None:
            self.transform = self.transform.fit(self.xs)

        # It doesnt matter if test or train, but test is usually smaller
        ts_info = parse_ts(path / ds_name / f"{ds_name}_TEST.ts")
        self.dimensions = ts_info.dimensions
        self.num_classes = ts_info.num_classes
        self.series_length = ts_info.series_length
        self.equal_length = ts_info.equal_length
        self.univariate = ts_info.univariate
        self.missing = ts_info.missing

        if raise_on_missing and ts_info.missing:
            raise MissingValueError(
                f"Dataset {ds_name} contains NaN values. If this is intended behavior, set `raise_on_missing=False`"
            )

        # convert string labels to int
        if self.ys.dtype.kind in {"U", "S"}:
            factorized_y = pd.factorize(self.ys, sort=True)
            self.y_labels = factorized_y[1]
            self.ys = factorized_y[0]

        self.ys = torch.tensor(self.ys, dtype=torch.long)

    def label_from_index(self, index: int) -> str:
        return self.y_labels[int(index)]

    def __len__(self) -> int:
        return len(self.xs)

    def __getitem__(self, index: int) -> tuple[torch.FloatTensor, torch.LongTensor]:
        if self.transform:
            return self.transform(self.xs[index]), self.ys[index]
        return self.xs[index], self.ys[index]



from pathlib import Path
from lightning import LightningDataModule

from torch.utils.data import DataLoader

from ..utils import swap_batch_seq_collate_fn

from ..transforms import Transform
from ..typing import DatasetSplit
from ..datasets.tfc_pretrain import TFCPretrainDataset, download_tfc_pretrain


class TFCPretrainDataModule(LightningDataModule):
    """See :class:`~torchchronos.datasets.TFCPretrainDataset` for more details."""

    def __init__(
        self,
        name: str,
        batch_size: int = 32,
        transform: Transform | None = None,
    ):
        super().__init__()

        self.name = name
        self.batch_size = batch_size
        self.transform = transform

        self.cache_dir = Path(".cache") / "data"

    def prepare_data(self) -> None:
        download_tfc_pretrain(self.cache_dir, self.name)

    def setup(self, stage: str | None = None) -> None:
        match stage:
            case "fit":
                self.train_dataset = TFCPretrainDataset(
                    self.name, self.cache_dir, DatasetSplit.TRAIN, self.transform
                )
                self.val_dataset = TFCPretrainDataset(
                    self.name, self.cache_dir, DatasetSplit.VAL, self.transform
                )
            case "validate":
                self.val_dataset = TFCPretrainDataset(
                    self.name, self.cache_dir, DatasetSplit.VAL, self.transform
                )
            case "test":
                self.test_dataset = TFCPretrainDataset(
                    self.name, self.cache_dir, "test", self.transform
                )
            case _:
                raise ValueError(f"Invalid stage: {stage}")

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            collate_fn=swap_batch_seq_collate_fn,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            collate_fn=swap_batch_seq_collate_fn,
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            collate_fn=swap_batch_seq_collate_fn,
        )



from pathlib import Path

import torch
from torch.utils.data import Dataset

from ..download import download_and_unzip_dataset
from ..transforms import Transform
from ..typing import DatasetSplit, AnyPath


class TFCPretrainDataset(Dataset):
    """Loads certain pieces of data from the TF-C paper [1]_ [2]_.

    Supported datasets:
        - "Gesture"
        - "EMG"

    They are preprocessed in the exact same way as in the original paper.

    The data will be downloaded to the directory specified by ``path`` if not already
    present. The data will be stored in a subdirectory named after the dataset.

    Args:
        name: Name of the dataset to load (see above for supported datasets).
        path: Path to the directory where the dataset should be stored.
        split: Which split of the dataset to load. Can be "train", "val" or "test".
        transform: A optional :class:`~Transform` to apply to the data.

    Shapes:
        The returned data is a tuple containing the data and the label. The data is a
        :class:`torch.FloatTensor` of shape ``(batch, sequence length, number of features)`` and the label is a
        :class:`torch.LongTensor` of shape ``(batch,)``.

    References:
        .. [1]
            Self-Supervised Contrastive Pre-Training For Time Series via Time-Frequency Consistency
            Xiang Zhang, Ziyuan Zhao, Theodoros Tsiligkaridis, and Marinka Zitnik
            Proceedings of Neural Information Processing Systems, NeurIPS 2022
            https://arxiv.org/abs/2206.08496.
        .. [2]
            Overview over the originally uploaded datasets on Figshare:
            https://figshare.com/search?q=%22Self-Supervised%20Contrastive%20Pre-Training%20For%20Time%20Series%20via%20Time-Frequency%20Consistency%22
    """

    NAME_TO_URL = {
        "Gesture": "https://figshare.com/ndownloader/articles/22634401/versions/2",
        "EMG": "https://figshare.com/ndownloader/articles/22634332/versions/2",
    }

    def __init__(
        self,
        name: str,
        path: AnyPath,
        split: DatasetSplit = DatasetSplit.TRAIN,
        transform: Transform | None = None,
    ) -> None:
        super().__init__()

        self.name = name
        self.path = path
        self.split = split
        self.transform = transform

        download_tfc_pretrain(self.path, self.name)
        self.full_path = Path(path) / name / f"{split.value}.pt"

        data = torch.load(self.full_path)
        self._samples = data["samples"].transpose(1, 2)  # move features to last dim
        self._labels = data["labels"].long()
        assert len(self._samples) == len(
            self._labels
        ), "Samples and labels must have the same length"

        if self.transform is not None:
            self.transform = self.transform.fit(self._samples)

    def __len__(self) -> int:
        return len(self._samples)

    def __getitem__(self, index: int) -> tuple[torch.FloatTensor, torch.LongTensor]:
        x = self._samples[index, ...]
        if self.transform is not None:
            x = self.transform(x)
        return x, self._labels[index]


def download_tfc_pretrain(path: AnyPath, name: str) -> None:
    dataset_path = Path(path) / name

    if not dataset_path.exists():
        url = TFCPretrainDataset.NAME_TO_URL[name]
        # This downloads all splits at once
        download_and_unzip_dataset(url, dataset_path)

    if not dataset_path.exists():
        raise RuntimeError(
            f"Downloaded data to {dataset_path.absolute()} but it does not exist."
        )



import torch


class Transform:
    def fit(self, ts: torch.Tensor) -> "Transform":
        return self

    def __call__(self, ts: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    def __repr__(self) -> str:
        raise NotImplementedError


class Compose(Transform):
    def __init__(self, transforms: list[Transform]):
        self.transforms = transforms

    def fit(self, ts: torch.Tensor) -> Transform:
        for t in self.transforms:
            t.fit(ts)
        return self

    def __call__(self, ts):
        for t in self.transforms:
            ts = t(ts)
        return ts

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string
